from .config import SQLALCHEMY_POOL_SIZE
from flask_sqlalchemy import SQLAlchemy


if SQLALCHEMY_POOL_SIZE is not None:
    db = SQLAlchemy(engine_options={'pool_size': int(SQLALCHEMY_POOL_SIZE)})
else:
    db = SQLAlchemy()



from sqlalchemy import select, and_, func, distinct
from sqlalchemy.orm import Session, aliased
from sqlalchemy.exc import NoResultFound
import logging as log
import os
from subprocess import run, Popen, PIPE, TimeoutExpired
import json as json_module
from datetime import datetime
from argparse import Namespace
import math
from time import sleep, time
from signal import SIGKILL
import re

from .model import Worker, Task, Execution, Job, Recruiter, Requirement, Signal, worker_delete, find_flavor, Flavor
from .config import WORKER_IDLE_CALLBACK, SERVER_CRASH_WORKER_RECOVERY, WORKER_OFFLINE_DELAY, WORKER_CREATE_CONCURRENCY,\
    WORKER_CREATE, WORKER_CREATE_RETRY, MAIN_THREAD_SLEEP, IS_SQLITE, SCITQ_SHORTNAME, TERMINATE_TIMEOUT, KILL_TIMEOUT,\
    JOB_MAX_LIFETIME
from .db import db
from ..util import PropagatingThread, to_obj, validate_protofilter
from ..constants import PROTOFILTER_SEPARATOR, PROTOFILTER_SYNTAX

protofilter_syntax=re.compile(PROTOFILTER_SYNTAX)

def get_nodename(session):
    worker_names = list(map(lambda x: x[0], 
        session.execute(select(Worker.name))))
    log.warning(f'Worker names: {worker_names}')
    i=1
    while f'{SCITQ_SHORTNAME}-node{i}' in worker_names:
        i+=1
    return f'{SCITQ_SHORTNAME}-node{i}'


def create_worker_object(concurrency, flavor, region, provider, batch, prefetch, db_session):
    """Create a worker object in db - this must be called linearly not in async way
    """
    hostname = get_nodename(db_session)
    log.info(f'Creating a new worker {hostname}: concurrency:{concurrency}, \
flavor:{flavor}, region:{region}, provider:{provider}, prefetch:{prefetch}')
    w = Worker(name=hostname, hostname=hostname, concurrency=concurrency, status='offline', 
            batch=batch, permanent=False, prefetch=prefetch, 
            flavor=flavor,region=region, provider=provider)
    db_session.add(w)
    db_session.commit()
    return w
    

def background(app):
    # while some tasks are pending without executions:
    #   look for a running worker:
    #      create a pending execution of this task for this worker


    with app.app_context():
        session = Session(db.engine)
    worker_process_queue = {}
    other_process_queue = []
    log.info('Starting thread for {}'.format(os.getpid()))
    ansible_workers = list(session.query(Worker).filter(and_(
                    Worker.status=='running',
                    ~Worker.permanent)).with_entities(
                        Worker.hostname))
    if ansible_workers:
        log.warning(f'Making sure workers {",".join([w.hostname for w in ansible_workers])} has access to server')
        process = PropagatingThread(
                    target=run,
                    args = (SERVER_CRASH_WORKER_RECOVERY,),
                    kwargs= {'shell': True, 'check':True}
                )
        other_process_queue.append(('Worker access task',process))
        process.start()

    while True:
        log.warning('Starting main loop')
        pending_job = None
        try:

            # ########    ###     ######  ##    ##    ########  ########   #######   ######  ########  ######   ######  #### ##    ##  ######   
            #    ##      ## ##   ##    ## ##   ##     ##     ## ##     ## ##     ## ##    ## ##       ##    ## ##    ##  ##  ###   ## ##    ##  
            #    ##     ##   ##  ##       ##  ##      ##     ## ##     ## ##     ## ##       ##       ##       ##        ##  ####  ## ##        
            #    ##    ##     ##  ######  #####       ########  ########  ##     ## ##       ######    ######   ######   ##  ## ## ## ##   #### 
            #    ##    #########       ## ##  ##      ##        ##   ##   ##     ## ##       ##             ##       ##  ##  ##  #### ##    ##  
            #    ##    ##     ## ##    ## ##   ##     ##        ##    ##  ##     ## ##    ## ##       ##    ## ##    ##  ##  ##   ### ##    ##  
            #    ##    ##     ##  ######  ##    ##    ##        ##     ##  #######   ######  ########  ######   ######  #### ##    ##  ######  


            task_list = list(session.query(Task).filter(
                    Task.status=='pending').with_entities(Task.task_id, Task.batch))
            if task_list:
                task_attributions = False
                worker_list = list(session.query(Worker).filter(
                            Worker.status=='running').with_entities(
                            Worker.worker_id,Worker.batch,Worker.concurrency,Worker.prefetch,Worker.task_properties))
                execution_per_worker_list = session.query(Execution.worker_id,Task.batch,Execution.status).\
                                                join(Execution.task).filter(and_(
                                                Execution.worker_id.in_(list([w.worker_id for w in worker_list])),
                                                Execution.status.in_(['running','pending','accepted']))
                                            )
                worker_properties = {worker.worker_id: json_module.loads(worker.task_properties) for worker in worker_list}
                execution_per_worker = {}
                for worker_id,batch,status in execution_per_worker_list:
                    if worker_id not in execution_per_worker:
                        execution_per_worker[worker_id] = 0
                    weight,_ = worker_properties[worker_id].get(batch, (1,0) )
                    if status=='running':
                        # only running task have a relative weight
                        execution_per_worker[worker_id] += weight
                    else:
                        execution_per_worker[worker_id] += 1
                
                for task in task_list:
                    for worker in worker_list:
                        if worker.batch != task.batch:
                            continue
                        if execution_per_worker.get(worker.worker_id,0)<(worker.concurrency+worker.prefetch):
                            session.add(Execution(worker_id=worker.worker_id,
                                task_id=task.task_id))
                            session.query(Task).filter(Task.task_id==task.task_id).update(
                                {'status':'assigned'}
                            )
                            execution_per_worker[worker.worker_id] = execution_per_worker.get(worker.worker_id,0)+1
                            log.info(f'Execution of task {task.task_id} proposed to worker {worker.worker_id}')
                            task_attributions = True
                            break
                if task_attributions:
                    session.commit()
            now = datetime.utcnow()

            # managing timeouts
            if IS_SQLITE:
                status_change_time = '(unixepoch(:now)-unixepoch(status_date))'
            else:
                # initially this was done with current_timestamp
                # but then SQLAlchemy with PostgreSQL issues the same answer for minutes (when it's supposed to change each second)
                # anyway, it works this way
                status_change_time = 'extract(epoch from :now - status_date)'
            some_timeouts = False
            log.warning(f'Looking for task timouts...')
            for item in session.execute(f'''SELECT execution.execution_id,execution.worker_id,execution.task_id 
                        FROM task 
                        JOIN execution ON execution.task_id=task.task_id AND execution.latest
                        WHERE 
                            (task.status='accepted' AND download_timeout IS NOT NULL AND {status_change_time}>download_timeout)
                            OR
                            (task.status='running' AND run_timeout IS NOT NULL AND {status_change_time}>run_timeout)''',params={'now':now}):
                log.warning(f'Task {item.task_id} has reached a timeout, sending kill signal.')
                session.add(Signal(execution_id=item.execution_id, worker_id=item.worker_id, signal=SIGKILL))
                some_timeouts = True
            if some_timeouts:
                session.commit()
            


            log.warning('Looking for offline/online workers')
            session.expire_all()
            for worker in session.query(Worker).filter(Worker.status.in_(['offline','running'])):
                change = False
                log.warning(f'Assessing worker {worker.name}')
                if worker.last_contact_date is None or (
                            now - worker.last_contact_date
                        ).total_seconds() > WORKER_OFFLINE_DELAY:
                    log.warning(f'... late: {worker.last_contact_date}')
                    if worker.status=='running':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) lost, marked as offline')
                        worker.status = 'offline'
                        change = True
                else:
                    log.warning('... good')
                    if worker.status=='offline':
                        log.warning(f'Worker {worker.name} ({worker.worker_id}) recovered, marked as running')
                        worker.status = 'running'
                        change = True
                if change:
                    session.commit()
            

            #        ##  #######  ########     ########  ########   #######   ######  ########  ######   ######  #### ##    ##  ######   
            #        ## ##     ## ##     ##    ##     ## ##     ## ##     ## ##    ## ##       ##    ## ##    ##  ##  ###   ## ##    ##  
            #        ## ##     ## ##     ##    ##     ## ##     ## ##     ## ##       ##       ##       ##        ##  ####  ## ##        
            #        ## ##     ## ########     ########  ########  ##     ## ##       ######    ######   ######   ##  ## ## ## ##   #### 
            #  ##    ## ##     ## ##     ##    ##        ##   ##   ##     ## ##       ##             ##       ##  ##  ##  #### ##    ##  
            #  ##    ## ##     ## ##     ##    ##        ##    ##  ##     ## ##    ## ##       ##    ## ##    ##  ##  ##   ### ##    ##  
            #   ######   #######  ########     ##        ##     ##  #######   ######  ########  ######   ######  #### ##    ##  ######     

            change = False
            for job in list(session.query(Job).filter(Job.status == 'pending')):
                pending_job = job

                if job.action == 'worker_destroy':
                    change=True
                    if ('destroy',job.target) in worker_process_queue:
                        log.warning(f'A destruction job is already running for worker {job.target}, failing this one')
                        job.status='failed'
                        job.log='Another destruction job is already running, this job failed as a doublon.'
                        continue
                    if ('create',job.target) in worker_process_queue:
                        worker,worker_create_process,job_id,start_time = worker_process_queue[('create',job.target)]
                        if worker_create_process.poll() is None:
                            worker_create_process.terminate()
                            try:
                                worker_create_process.wait(timeout=TERMINATE_TIMEOUT)
                                log.warning(f'Worker {job.target} creation process has been terminated')
                            except TimeoutExpired:
                                worker_create_process.kill()
                                try:
                                    worker_create_process.wait(timeout=KILL_TIMEOUT)
                                    log.warning(f'Worker {job.target} creation process has been killed')
                                except TimeoutExpired:
                                    log.exception(f'Could not kill worker {job.target} creation process, giving up for this round!')
                                    continue
                            
                            del(worker_process_queue[('create',job.target)])
                            session.query(Job).get(job_id).status='failed'
                    worker = Namespace(**job.args)
                    real_worker = session.query(Worker).get(worker.worker_id)
                    if real_worker.ansible_active:
                        if len(worker_process_queue)<WORKER_CREATE_CONCURRENCY:
                            worker_destroy_command = WORKER_IDLE_CALLBACK.format(hostname=job.target)
                            log.warning(f'Launching destroy process for {job.target}, command is "{worker_destroy_command}"')
                            worker_delete_process = Popen(
                                    worker_destroy_command,
                                    stdout = PIPE,
                                    stderr = PIPE,
                                    shell = True,
                                    encoding = 'utf-8'
                                )
                            job.status='running'
                            worker_process_queue[('destroy',job.target)]=(worker, worker_delete_process, job.job_id, time())
                            log.warning(f'Worker {job.target} destruction process has been launched')
                    else:
                        #TODO: this should be a very rare event, we should do extra tests here
                        log.warning(f'Deleting worker {worker.name} ({worker.worker_id})')
                        if real_worker is not None:
                            change = True
                            worker_delete(real_worker, session, is_destroyed=True, commit=False)
                            #session.delete(real_worker)
                            job.log='Deleting unmanaged worker.'
                            job.status='succeeded'            
                
                if job.action == 'worker_create':
                    change = True
                    worker = create_worker_object(db_session=session,
                        **job.args)
                    
                    auto_deploy = False
                    if worker.provider=='auto' or worker.region=='auto' or worker.flavor.startswith('auto'):
                        try:
                            log.warning(' -> Auto deploy detected')
                            auto_deploy = True
                            provider = None if worker.provider=='auto' else worker.provider
                            region = None if worker.region=='auto' else worker.region
                            protofilters=''
                            if worker.flavor.startswith('auto'):
                                validate_protofilter(worker.flavor)
                                flavor=None
                                if PROTOFILTER_SEPARATOR in worker.flavor:
                                    protofilters = PROTOFILTER_SEPARATOR.join(
                                            worker.flavor.split(PROTOFILTER_SEPARATOR)[1:])
                            else:
                                flavor=worker.flavor
                            flavor_list=list(map(to_obj,find_flavor(session, provider=provider, region=region, 
                                                                flavor=flavor, protofilters=protofilters, limit=None)))
                            while flavor_list:
                                current_flavor=flavor_list[0]
                                if current_flavor.available is not None and current_flavor.available<=0:
                                    flavor_list.pop(0)
                                    continue
                                else:
                                    worker.flavor=current_flavor.name
                                    worker.region=current_flavor.region
                                    worker.provider=current_flavor.provider
                                    break
                            else:
                                raise RuntimeError(f'Could not find a flavor satisfying provider={worker.provider},region={worker.region},flavor={worker.flavor}')
                        except RuntimeError as re:
                            job.status='failed'
                            job.log=re.args[0]

                    if job.status!='failed':
                        job.action = 'worker_deploy'
                        job.target = worker.name
                        job.args = dict(job.args)
                        job.args['worker_id'] = worker.worker_id
                        if auto_deploy:
                            job.args['flavor'] = worker.flavor
                            job.args['provider'] = worker.provider
                            job.args['region'] = worker.region
                        job.retry = WORKER_CREATE_RETRY
                        job.status = 'pending'
                
                if job.action == 'worker_deploy':
                    if ('create',job.target) not in worker_process_queue and len(
                                worker_process_queue)<WORKER_CREATE_CONCURRENCY:
                        if ('destroy',job.target) in worker_process_queue:
                            log.warning(f'Trying to recreate worker {job.target} after destruction too soon, waiting a little bit...')
                            continue
                        change = True
                        log.warning(f'Launching creation process for worker {job.target}.')
                        worker = Namespace(**job.args)
                        try:
                            flavor = session.query(Flavor).filter(Flavor.provider==worker.provider, 
                                                                Flavor.name==worker.flavor).one()
                        except NoResultFound:
                            log_message = f'Could not find flavor {worker.flavor} in {worker.provider}: worker {job.target} deploy failed.'
                            log.exception(log_message)
                            job.status='failed'
                            job.log=log_message
                            session.query(Worker).filter(Worker.worker_id==worker.worker_id).update(
                                {'status':'failed'}
                            )
                            continue
                        log.exception(f'Launching command is "'+WORKER_CREATE.format(
                            hostname=job.target,
                            concurrency=worker.concurrency,
                            flavor=worker.flavor,
                            region=worker.region,
                            provider=worker.provider,
                            tags=flavor.tags
                        )+'"')
                        worker_create_process = Popen(
                            WORKER_CREATE.format(
                                hostname=job.target,
                                concurrency=worker.concurrency,
                                flavor=worker.flavor,
                                region=worker.region,
                                provider=worker.provider,
                                tags=flavor.tags
                            ),
                            stdout = PIPE,
                            stderr = PIPE,
                            shell = True,
                            encoding = 'utf-8'
                        )
                        worker.name = worker.hostname = job.target
                        job.status = 'running'
                        worker_process_queue[('create',job.target)]=(worker, worker_create_process, job.job_id, time())
                        log.warning(f'Worker {job.target} creation process has been launched')

            if change:
                session.commit()                
            pending_job = None

            change = False
            for ((action,worker_name),(worker,worker_process,job_id,start_time)) in list(worker_process_queue.items()):
                returncode = worker_process.poll()
                if returncode is not None:
                    change = True
                    job = session.query(Job).get(job_id)
                    del(worker_process_queue[(action,worker_name)])
                    if returncode == 0:
                        log.warning(f'Process {action} succeeded for worker {worker.name}.')
                        job.log = worker_process.stdout.read()
                        job.status = 'succeeded'

                        if action=='destroy':
                            #session.execute(Worker.__table__.delete().where(
                            #    Worker.__table__.c.worker_id==worker.worker_id))
                            log.warning(f'Deleting worker {worker.name} ({worker.worker_id}) after destruction')
                            real_worker = session.query(Worker).get(worker.worker_id)
                            if real_worker is not None:
                                #session.delete(real_worker)
                                worker_delete(real_worker, session, is_destroyed=True, commit=False)
                            else:
                                log.error(f'Could not find a worker with worker_id {worker.worker_id}')
                            #worker_dao.delete(worker.worker_id, is_destroyed=True)
                    else:
                        stderr = worker_process.stderr.read()
                        log.warning(f'Process {action} failed for worker {worker.name}: {stderr}')
                        job.log = worker_process.stdout.read() + stderr
                        log.warning(f'Job output is {job.log}')
                        if job.retry > 0:
                            job.retry -= 1
                            job.status = 'pending'
                        else:
                            job.status = 'failed'
                            if action=='create':
                                worker = session.query(Worker).get(job.args['worker_id'])
                                worker.status = 'failed'
                elif time() - start_time > JOB_MAX_LIFETIME:
                    log.warning(f'Process {action} is taking too long for worker {worker.name}, sending TERM signal.')
                    worker_process.terminate()
                elif time() - start_time > JOB_MAX_LIFETIME + TERMINATE_TIMEOUT:
                    log.warning(f'Process {action} is really taking too long for worker {worker.name}, sending KILL signal.')
                    worker_process.kill()

            for job in list(session.query(Job).filter(Job.status == 'running')):
                if job.action=='worker_deploy':
                    action='create'
                elif job.action=='worker_destroy':
                    action='destroy'
                else:
                    action=job.action
                if (action, job.target) not in worker_process_queue:
                    log.warning(f'Job {(job.action, job.target)} seems to have failed, not in {worker_process_queue}')
                    job.status='failed'
                    change = True

            if change:
                session.commit()
                
            for process_name, process in list(other_process_queue):
                try:
                    if not process.is_alive():
                        other_process_queue.remove((process_name, process))
                        log.warning(f'Job {process_name} is done.')
                except Exception as e:
                    log.exception(f'Job {process_name} failed: {e}')
                    other_process_queue.remove((process_name, process))

                                
            ########  ########  ######  ########  ##     ## #### ######## ######## ########   ######  
            ##     ## ##       ##    ## ##     ## ##     ##  ##     ##    ##       ##     ## ##    ## 
            ##     ## ##       ##       ##     ## ##     ##  ##     ##    ##       ##     ## ##       
            ########  ######   ##       ########  ##     ##  ##     ##    ######   ########   ######  
            ##   ##   ##       ##       ##   ##   ##     ##  ##     ##    ##       ##   ##         ## 
            ##    ##  ##       ##    ## ##    ##  ##     ##  ##     ##    ##       ##    ##  ##    ## 
            ##     ## ########  ######  ##     ##  #######  ####    ##    ######## ##     ##  ######  
                        
            change = False
            task1 = aliased(Task)
            task2 = aliased(Task)
            recyclable_worker_active_tasks = {}
            worker_task_properties = {}
            worker_active_batch = {}
            recyclable_worker_list = list(session.query(Worker,Task.batch,func.count(distinct(Task.task_id))).\
                    join(Execution,and_(Worker.worker_id==Execution.worker_id,Execution.status=='running'), isouter=True).\
                    join(Execution.task,isouter=True).group_by(Worker,Task.batch))
            worker_batch_task = dict(session.query(Worker,func.count(Task.task_id)).\
                                     join(Task, and_(Task.batch==Worker.batch, Task.status=='pending'),isouter=True).\
                                     group_by(Worker))
            for worker,batch,worker_tasks in recyclable_worker_list:
                if worker not in worker_batch_task or worker_batch_task[worker] == 0:
                    if worker not in recyclable_worker_active_tasks:
                        recyclable_worker_active_tasks[worker]=0
                        worker_task_properties[worker] = json_module.loads(worker.task_properties)
                        worker_active_batch[worker]=[]
                    weight,_ = worker_task_properties[worker].get(batch,(1,0))
                    recyclable_worker_active_tasks[worker] += weight * worker_tasks
            
                    if worker_tasks == 0:
                        if batch in worker_task_properties[worker]:
                            log.warning(f'-> Cleaning batch {batch} from worker {worker} task properties 1')
                            change=True
                            del worker_task_properties[worker][batch] 
                            worker.task_properties=json_module.dumps(worker_task_properties[worker])
                    else:
                        worker_active_batch[worker].append(batch)
            for worker,task_properties in worker_task_properties.items():
                for batch in list(task_properties.keys()):
                    if batch not in worker_active_batch[worker]:
                        log.warning(f'-> Cleaning batch {batch} from worker {worker} task properties 2 : {worker_task_properties}')
                        change=True
                        del worker_task_properties[worker][batch]
                        worker.task_properties=json_module.dumps(worker_task_properties[worker])
            active_recruiters = list(session.query(Recruiter,func.count(distinct(Task.task_id)),func.count(distinct(Worker.worker_id))).\
                    join(Task,and_(Task.batch==Recruiter.batch,Task.status=='pending')).\
                    join(Worker,Worker.batch==Recruiter.batch,isouter=True).\
                    group_by(Recruiter.batch,Recruiter.rank).order_by(Recruiter.batch,Recruiter.rank))
            #pending_workers = dict(session.query(Recruiter,func.count(distinct(Worker.worker_id))).\
            #        join(Task,and_(Task.batch==Recruiter.batch,Task.status=='pending')).\
            #        join(Worker,and_(Worker.batch==Recruiter.batch,Worker.status!='running'),isouter=True).\
            #        group_by(Recruiter.batch,Recruiter.rank).order_by(Recruiter.batch,Recruiter.rank))
            newly_recruited_workers = []   
            for recruiter,pending_tasks,workers in active_recruiters:
                log.warning(f'-> recruiting for recruiter {recruiter} with {pending_tasks} pending tasks and {workers} current workers')
                if recruiter.minimum_tasks and recruiter.minimum_tasks > pending_tasks:
                    log.warning(f'  --> not enough tasks, not recruiting ({pending_tasks} is below the minimum of {recruiter.minimum_tasks})')
                    continue
                if recruiter.maximum_workers and recruiter.maximum_workers <= workers:
                    log.warning(f'  --> too many workers already, not recruiting ({recruiter} has reached the maximum of {recruiter.maximum_workers})')
                    continue
                nb_workers = math.ceil(pending_tasks/recruiter.tasks_per_worker) - workers
                log.warning(f'  --> we need {nb_workers} because {pending_tasks} pending tasks ({recruiter.tasks_per_worker} expected per worker) and we already have {workers} workers.')
                if recruiter.maximum_workers and recruiter.maximum_workers < nb_workers + workers:
                    nb_workers = recruiter.maximum_workers - workers
                    log.warning(f'  --> adjusting to {nb_workers} because maximum is {recruiter.maximum_workers} and \
we already have {workers} workers')
                if nb_workers <= 0:
                    log.warning(f'  --> giving up')
                    continue
                
                if len(recyclable_worker_active_tasks.items())==0:
                    log.warning(f'  --> No recyclable workers {recyclable_worker_active_tasks}')
                for worker, active_tasks  in recyclable_worker_active_tasks.items():  
                    if worker.batch != recruiter.batch and recruiter.match_flavor(worker, session=session) \
                            and active_tasks<worker.concurrency and nb_workers>0 and worker not in newly_recruited_workers:
                        previous_batch = worker.batch
                        previous_concurrency = worker.concurrency
                        previous_task_properties = worker_task_properties[worker]

                        task_ratio = recruiter.worker_concurrency / previous_concurrency
                        task_properties = dict([(batch, ( max(ratio*task_ratio,recruiter.worker_concurrency) , prio+1 )) 
                                                for batch,(ratio,prio) in previous_task_properties.items()])
                        task_properties[previous_batch] = (task_ratio,1)

                        if recruiter.batch in task_properties:
                            log.warning(f'-> Recruiter {recruiter} tried to recycle {worker} but some task of the same batch from a previous recycling round are present, giving up')
                            continue
                        
                        log.warning(f'-> Recruiting worker {worker} from batch {previous_batch} to {recruiter.batch}')
                        newly_recruited_workers.append(worker)
                        change = True
                        worker.batch = recruiter.batch
                        worker.prefetch = recruiter.worker_prefetch
                        worker.concurrency = recruiter.worker_concurrency
                        worker.task_properties = json_module.dumps(task_properties)
                        session.add(worker)
                        nb_workers -= 1
                    else:
                        if worker.batch == recruiter.batch:
                            reason = 'already recruited'
                        elif worker.flavor != recruiter.worker_flavor:
                            reason = 'not the right flavor'
                        elif active_tasks>=worker.concurrency:
                            reason = f'too busy {active_tasks}>={worker.concurrency}'
                        elif nb_workers <= 0:
                            reason = f'no more workers needed'
                        elif worker in newly_recruited_workers:
                            reason = f'it was recently recruited by another recruiter'
                        else:
                            reason = f'something mysterious that should never occur'
                        log.warning(f'  --> {worker} not suitable because {reason} ')
                
                if recruiter.worker_provider is not None and recruiter.worker_region is not None and nb_workers>0:
                    if recruiter.worker_provider=='auto' or recruiter.worker_region=='auto' or recruiter.worker_flavor.startswith('auto'):
                        log.warning(' -> Auto recruiter detected')
                        worker_provider = None if recruiter.worker_provider=='auto' else recruiter.worker_provider
                        worker_region = None if recruiter.worker_region=='auto' else recruiter.worker_region
                        protofilters=''
                        if recruiter.worker_flavor.startswith('auto'):
                            worker_flavor=None
                            if PROTOFILTER_SEPARATOR in recruiter.worker_flavor:
                                protofilters = PROTOFILTER_SEPARATOR.join(
                                        recruiter.worker_flavor.split(PROTOFILTER_SEPARATOR)[1:])
                        else:
                            worker_flavor=recruiter.worker_flavor
                        worker_to_find=nb_workers
                        
                        while worker_to_find>0:
                            flavor_list=find_flavor(session, provider=worker_provider, region=worker_region, 
                                                            flavor=worker_flavor, protofilters=protofilters, limit=None)
                            for current_flavor in flavor_list:
                                if current_flavor['available']>0:
                                    break
                            else:
                                break
                            worker_to_find-=1
                            current_flavor=to_obj(current_flavor)
                            log.warning(f'-> Deploying one worker from {current_flavor.provider},{current_flavor.region} : {current_flavor.name}')
                            session.add(
                                Job(target='', 
                                    action='worker_create', 
                                    args={
                                        'concurrency': recruiter.worker_concurrency, 
                                        'prefetch': recruiter.worker_prefetch,
                                        'flavor': current_flavor.name,
                                        'region': current_flavor.region,
                                        'provider': current_flavor.provider,
                                        'batch': recruiter.batch
                                    }
                                )
                            )
                            change = True
                        if worker_to_find>0:
                            log.warning(f'-> Could not find enough available flavors of the right kind, missing {worker_to_find}')
                            
                                


                    else:
                        for _ in range(nb_workers):
                            log.warning(f'-> Deploying one worker from {recruiter.worker_provider}')
                            session.add(
                                Job(target='', 
                                    action='worker_create', 
                                    args={
                                        'concurrency': recruiter.worker_concurrency, 
                                        'prefetch': recruiter.worker_prefetch,
                                        'flavor': recruiter.worker_flavor,
                                        'region': recruiter.worker_region,
                                        'provider': recruiter.worker_provider,
                                        'batch': recruiter.batch
                                    }
                                )
                            )
                        change = True
                
            if change:
                session.commit()
                



            ########  ########  #######  ##     ## #### ########  ######## ##     ## ######## ##    ## ########  ######  
            ##     ## ##       ##     ## ##     ##  ##  ##     ## ##       ###   ### ##       ###   ##    ##    ##    ## 
            ##     ## ##       ##     ## ##     ##  ##  ##     ## ##       #### #### ##       ####  ##    ##    ##       
            ########  ######   ##     ## ##     ##  ##  ########  ######   ## ### ## ######   ## ## ##    ##     ######  
            ##   ##   ##       ##  ## ## ##     ##  ##  ##   ##   ##       ##     ## ##       ##  ####    ##          ## 
            ##    ##  ##       ##    ##  ##     ##  ##  ##    ##  ##       ##     ## ##       ##   ###    ##    ##    ## 
            ##     ## ########  ##### ##  #######  #### ##     ## ######## ##     ## ######## ##    ##    ##     ######  



            change = False
            task1 = aliased(Task)
            task2 = aliased(Task)
            active_requirements = session.query(task1,task2.status).\
                                    join(task1,Requirement.task).filter(task1.status == 'waiting').\
                                    join(task2,Requirement.other_task)
            candidate_tasks = []
            failed_requirements = []
            for task,status in list(active_requirements):
                if status=='succeeded' and task not in candidate_tasks:
                    candidate_tasks.append(task)
                if status!='succeeded' and task not in failed_requirements:
                    failed_requirements.append(task)
            for task in candidate_tasks:
                if task not in failed_requirements:
                    task.status='pending'
                    session.add(task)
                    change = True
            if change:
                session.commit()




                
        except Exception as e1:
            log.exception('An exception occured during server main loop:')
            while True:
                sleep(MAIN_THREAD_SLEEP)
                try:
                    log.warning('Trying to reconnect...')
                    session.close()
                    with app.app_context():
                        session = Session(db.engine)
                    break
                except Exception as e2:
                    log.exception(f'Error while trying to reconnect: {e2}')
            if pending_job is not None:
                session.query(Job).filter(Job.job_id==job.job_id).update(
                    {'status':'failed', 'log':f'Job failed due to exception {e1}, see logs for details'})
                session.commit()
        sleep(MAIN_THREAD_SLEEP)


import ovh
from getpass import getpass
from argparse import Namespace
import pandas as pd
import requests
import re
from ...server import config, get_session
from ..generic import Flavor, FlavorMetrics, GenericProvider
from ...util import to_obj
import logging as log

# Instantiate an OVH Client.
#print('''If you don't have credentials you can create ones with full access to your account on the token creation page:
#       https://api.ovh.com/createToken/index.cgi?GET=/*&PUT=/*&POST=/*&DELETE=/*
#''')

OVH_ENDPOINT='ovh-eu'

class OVH(GenericProvider):

    def __init__(self, session):
        self.session = session

        if config.OVH_APPLICATIONKEY == '' \
            or config.OVH_APPLICATIONSECRET == '' \
            or config.OVH_CONSUMERKEY == '':
            raise RuntimeError('''OVH_APPLICATIONKEY, OVH_APPLICATIONSECRET and OVH_CONSUMERKEY *must* be set in /etc/scitq.conf 
(or as shell variables) for the updater to work.
You may have to visit the following website to create them:
 https://api.ovh.com/createToken/index.cgi?GET=/* ''')
        
        if config.OVH_REGIONS == '':
            raise RuntimeError('''OVH_REGIONS *must* be set in /etc/scitq.conf for the updater to work.''')

        self.client = ovh.Client(
            endpoint=OVH_ENDPOINT,               # Endpoint of API OVH (List of available endpoints: https://github.com/ovh/python-ovh#2-configure-your-application)
            application_key=config.OVH_APPLICATIONKEY,    # Application Key
            application_secret=config.OVH_APPLICATIONSECRET, # Application Secret
            consumer_key=config.OVH_CONSUMERKEY,       # Consumer Key
        )
        self.provider = 'ovh'
        self.service_name = config.OS_PROJECT_ID
        self.regions = config.OVH_REGIONS.split() if config.OVH_REGIONS else \
            self.client.get(f"/cloud/project/{self.service_name}/region")
        self.flavors = {}
        self.flavor_metrics = {}
        self.live = True

    #result = client.get('/cloud/project')
    #choice = None
    #while choice is None or not(0<choice<=len(result)):
    #    print('Choose your service name between: ')
    #    for i,project in enumerate(result):
    #        print(f'choice {i+1}) {project}')
    #    choice=int(input(f'Answer between 1 and {len(result)}: '))
    #service_name = result[choice-1]

    def get_flavors(self):
        """Update OVH flavors"""
        raw_flavors = self.client.get(f"/cloud/project/{self.service_name}/flavor")
        self.push('Getting flavors')
        for flavor in map(to_obj, raw_flavors):
            if not flavor.available or flavor.osType=='windows' or 'flex' in flavor.name:
                continue
            self.push('.')
            if flavor.name not in self.flavors:
                tags=''
                if flavor.name.startswith('bm'):
                    tags+='M'
                if flavor.name=='t1-90' and flavor.vcpus==16:
                    # for some strange reason OVH API sometimes (but not always) answer wrong here...
                    flavor.vcpus=18
                    self.push('<hack t1-90>')
                try:
                    f=Flavor(
                        name=flavor.name,
                        provider=self.provider,
                        cpu=flavor.vcpus,
                        ram=flavor.ram/1000,
                        disk=flavor.disk,
                        bandwidth=flavor.inboundBandwidth/1000,
                        tags=tags)
                except:
                    log.exception('Flavor ill created')
                    self.push('!')
                self.flavors[flavor.name]=f
            if (flavor.name, flavor.region) not in self.flavor_metrics:
                try:
                    fm=FlavorMetrics(
                        flavor_name=flavor.name,
                        provider=self.provider,
                        region_name=flavor.region,
                        eviction=0)
                except:
                    log.exception('FlavorMetrics ill created')
                    self.push('!')
                self.flavor_metrics[(flavor.name, flavor.region)]=fm
        self.push('\n')

    def get_metrics(self):
        self.push('Getting metrics')
        if not self.flavors:
            raise RuntimeError('Cannot get metrics on an empty flavor set')
        
        r = requests.get('https://www.ovhcloud.com/fr/public-cloud/prices/')
        for table in pd.read_html(r.content):
            for line in table.itertuples():
                if hasattr(line,'Nom') and line.Nom in self.flavors:
                    self.push('.')
                    flavor = line.Nom
                    
                    # cost
                    m=re.match(r'([0-9,]+).€ HT/heure',line.Prix)
                    if m:
                        for region in self.regions:
                            if (flavor,region) in self.flavor_metrics:
                                try:
                                    self.flavor_metrics[(flavor,region)].cost = float(m.groups()[0].replace(',','.'))
                                except:
                                    log.exception('FlavorMetrics cost could not be completed')
                                    self.push('!')
                    else:
                        self.push('!')
                        log.exception(f'Flavor {flavor}, could not understand cost: '+line.Prix)

                    # disk
                    disk_found = False
                    for m in re.finditer(r'.*?((?P<n>[0-9]+).x.)?(?P<value>[0-9.]+).(?P<unit>T|G)o.*?',line.Stockage):
                        disk_found = True
                        dict = m.groupdict()
                        n=dict['n']
                        value=dict['value']
                        unit=dict['unit']
                        n=1 if n is None else int(n)
                        value=float(value)
                        unit=1 if unit=='G' else 1024
                        disk = n*value*unit
                        try:
                            if not self.flavors[flavor].disk or self.flavors[flavor].disk<disk:
                                self.flavors[flavor].disk=disk
                        except:
                            log.exception('Flavor disk could not be completed')
                            self.push('!')
                    if not disk_found:
                        self.push('!')
                        log.exception(f'Flavor {flavor}, could not understand disk: '+line.Stockage)
                    
                    if 'NVMe' in line.Stockage:
                        try:
                            self.flavors[flavor].tags+='N'
                        except:
                            log.exception('Flavor tag could not be completed')
                            self.push('!')
                    
                    if hasattr(line,'GPU'):
                        try:
                            self.flavors[flavor].tags+='G'
                            self.flavors[flavor].gpu=line.GPU
                            m=re.match(r'((?P<n>[0-9]+).)?.+ (?P<value>[0-9]+).Go',line.GPU)
                            if m:
                                dicts=m.groupdict()
                                n=1 if dicts['n'] is None else int(dicts['n'])
                                value=float(dicts['value'].replace(',','.'))
                                self.flavors[flavor].gpumem=n*value
                        except:
                            log.exception('Flavor gpu info could not be completed')
                            self.push('!')

        self.push('\n')
        self.update_flavors(self.flavors.values())
        self.update_flavor_metrics([fm for fm in self.flavor_metrics.values() if fm.cost])
        self.push('done\n')

def run():
    session = get_session()

    updater=OVH(session=session)
    updater.get_flavors()
    updater.get_metrics()


from sqlalchemy import and_
from sqlalchemy.orm.exc import NoResultFound
from ..server.model import Flavor, FlavorMetrics
from io import StringIO

import logging as log

def show_key(primary_keys, premary_key_values):
    return ','.join([f'{k}:{v}' for k,v in zip(primary_keys, premary_key_values)])

class GenericProvider:

    def __init__(self, session, provider):
        self.session = session
        self.provider = provider

    def push(self, x, cache=StringIO()):
        """A small wrapper on print with options"""
        if self.live:
            print(x, end='', flush=True)
        else:
            cache.write(x)
            if '\n' in x:
                log.warning(cache.getvalue())
                cache.seek(0)
                cache.truncate()

    def update_flavors(self, flavors):
        """Update an existing flavor or create a new one"""
        self.generic_update(object_class=Flavor,
                            current=flavors,
                            provider=self.provider,
                            attributes=['cpu','ram','disk','bandwidth', 'gpu', 'gpumem', 'tags'],
                            primary_keys=['provider','name'])

    def update_flavor_metrics(self, metrics):
        """Update some flavor metrics (e.g. regional details for a flavor)"""
        self.generic_update(object_class=FlavorMetrics,
                            current=metrics,
                            provider=self.provider,
                            attributes=['eviction','cost'],
                            primary_keys=['provider','flavor_name','region_name'])

    def generic_update(self, object_class, current, provider, attributes, primary_keys):
        """Compare a list of object of class object_class with its current content in base and update"""
        current_dict = { tuple ( (getattr(object,pk) for pk in primary_keys) ) : object for object in current }
        for object in self.session.query(object_class).filter(object_class.provider==provider).all():
            pkey = tuple( (getattr(object,pk) for pk in primary_keys) )
            if pkey not in current_dict:
                self.push(f'{object_class.__name__} {show_key(primary_keys,pkey)} is removed\n')
                self.session.delete(object)
            else:
                new_object = current_dict.pop(pkey)
                for attr in attributes:
                    val=getattr(object,attr)
                    new_val=getattr(new_object,attr)
                    if val!=new_val:
                        self.push(f'{object_class.__name__} {show_key(primary_keys,pkey)} update {attr}: {val}->{new_val}\n')
                        setattr(object, attr, new_val)
        for _,object in current_dict.items():
            pkey = tuple( (getattr(object,pk) for pk in primary_keys) )
            self.push(f'new {object_class.__name__} {show_key(primary_keys,pkey)}\n')
            self.session.add(object)
        self.session.commit()





from datetime import datetime
import json as json_module
from sqlalchemy import DDL, event, func, delete, select, or_, and_, tuple_
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import func

from .config import DEFAULT_BATCH, WORKER_DESTROY_RETRY, get_quotas, EVICTION_ACTION, EVICTION_COST_MARGIN, PREFERRED_REGIONS
from .db import db
from ..util import to_dict, validate_protofilter, protofilter_syntax, PROTOFILTER_SEPARATOR, is_like, has_tag
from ..constants import FLAVOR_DEFAULT_EVICTION, FLAVOR_DEFAULT_LIMIT, EXECUTION_STATUS, WORKER_STATUS

import logging as log

class ModelException(Exception):
    @property
    def message(self):
        return self.args[0]

class Task(db.Model):
    __tablename__ = "task"
    task_id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String)
    command = db.Column(db.String, nullable=False)
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    batch = db.Column(db.String, nullable=False, default=DEFAULT_BATCH)
    input = db.Column(db.String, nullable=True)
    output = db.Column(db.String, nullable=True)
    container = db.Column(db.String, nullable=True)
    container_options = db.Column(db.String, nullable=True)
    resource = db.Column(db.String, nullable=True)
    retry = db.Column(db.Integer, nullable=False, default=0)
    status_date = db.Column(db.DateTime)
    download_timeout = db.Column(db.Integer, nullable=True)
    run_timeout = db.Column(db.Integer, nullable=True)

    def __init__(self, command, name=None, status='pending', batch=None, 
                    input=None, output=None, container=None, 
                    container_options=None, resource=None,
                    download_timeout=None, run_timeout=None,
                    retry=None):
        self.name = name
        self.command = command
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.status_date = self.creation_date
        self.batch = batch if batch is not None else DEFAULT_BATCH
        if '/' in self.batch:
            raise ModelException(f'Cannot accept / in batch name, chose a proper batch name: {self.batch}')
        self.input = input
        self.output = output
        self.container = container
        self.container_options = container_options
        self.resource = resource
        self.retry = retry
        self.download_timeout = download_timeout
        self.run_timeout = run_timeout


class Worker(db.Model):
    __tablename__ = "worker"
    worker_id = db.Column(db.Integer,primary_key=True)
    name = db.Column(db.String, nullable=False)
    hostname = db.Column(db.String)
    status = db.Column(db.String, nullable=False)
    concurrency = db.Column(db.Integer,nullable=False)
    prefetch = db.Column(db.Integer,default=0)
    load = db.Column(db.String)
    memory=db.Column(db.String)
    stats=db.Column(db.String)
    task_properties=db.Column(db.String, default=json_module.dumps({}))
    creation_date = db.Column(db.DateTime)
    modification_date = db.Column(db.DateTime)
    last_contact_date = db.Column(db.DateTime)
    batch = db.Column(db.String, nullable=False,default=DEFAULT_BATCH)
    permanent = db.Column(db.Boolean, default=True)
    flavor = db.Column(db.String, nullable=True)
    region = db.Column(db.String, nullable=True)
    provider = db.Column(db.String, nullable=True)
    ipv4 = db.Column(db.String, nullable=True)
    ipv6 = db.Column(db.String, nullable=True)
    ansible_host = db.Column(db.String, nullable=True)
    ansible_group = db.Column(db.String, nullable=True)
    ansible_active = db.Column(db.Boolean, default=False)
    signals = db.relationship("Signal", cascade="all,delete")

    def __init__(self, name, concurrency, prefetch=0, hostname=None, 
                status='paused', batch=None, flavor=None, region=None, provider=None, permanent=True):
        self.name = name
        self.concurrency = concurrency
        self.prefetch = prefetch
        self.status = status
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.hostname = hostname
        self.batch = batch
        self.permanent = permanent
        self.region = region
        self.flavor = flavor
        self.provider = provider
     

class Execution(db.Model):
    __tablename__ = "execution"
    execution_id = db.Column(db.Integer, primary_key=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"), nullable=True)
    worker = db.relationship(
        Worker,
        backref=db.backref('executions',
                         uselist=True,
                         cascade='save-update',
                         order_by='Execution.modification_date.desc()'))
    task_id = db.Column(db.Integer, db.ForeignKey("task.task_id"), nullable=False)
    task = db.relationship(
        Task,
        backref=db.backref('executions',
                         uselist=True,
                         cascade='delete,all'),
                         order_by='Execution.creation_date')
    status = db.Column(db.String, nullable=False)
    creation_date = db.Column(db.DateTime, nullable=False)
    modification_date = db.Column(db.DateTime)
    output = db.Column(db.Text)
    error = db.Column(db.Text)
    return_code = db.Column(db.Integer)
    pid = db.Column(db.String)
    output_files = db.Column(db.String, nullable=True)
    command = db.Column(db.String, nullable=True)
    latest = db.Column(db.Boolean, default=True)
    

    def __init__(self, worker_id, task_id, status='pending', pid=None, 
                    return_code=None, command=None):
        self.worker_id = worker_id
        self.task_id = task_id
        self.status = status
        self.pid = pid
        self.return_code = return_code
        self.creation_date = datetime.utcnow()
        self.modification_date = self.creation_date
        self.command = command

trigger_latest_sqlite = DDL("""
        CREATE TRIGGER is_latest BEFORE INSERT ON execution FOR EACH ROW 
        BEGIN
            UPDATE execution SET latest=false WHERE latest AND task_id=NEW.task_id;
        END
        """)

func_latest_postgres = DDL("""
        CREATE OR REPLACE FUNCTION render_obsolete() 
        RETURNS TRIGGER
        LANGUAGE plpgsql
        AS $$
        BEGIN
            UPDATE execution 
            SET latest=false
            WHERE task_id=NEW.task_id
            AND latest;
            RETURN NEW;
        END;
        $$
        """)

trigger_latest_postgres = DDL("""
CREATE TRIGGER is_latest BEFORE INSERT ON execution FOR EACH ROW EXECUTE PROCEDURE render_obsolete()
""")

event.listen(
    Execution.__table__, 'after_create',
    trigger_latest_sqlite.execute_if(dialect="sqlite")    
)
event.listen(
    Execution.__table__, 'after_create',
    func_latest_postgres.execute_if(dialect="postgresql")    
)
event.listen(
    Execution.__table__, 'after_create',
    trigger_latest_postgres.execute_if(dialect="postgresql")    
)

def execution_update_status(execution, session, status, commit=True):
    """Change status of an execution, impacting on task if required"""
    if status not in EXECUTION_STATUS:
        raise ModelException(f"Status {status} is not possible (only {' '.join(EXECUTION_STATUS)})")
    task=execution.task
    now = datetime.utcnow()
    if execution.status=='pending':
        if status=='running':
            task.status = 'running'
            task.modification_date = now
            task.status_date = now
        elif status=='accepted':
            task.status = 'accepted'
            task.modification_date = now
            task.status_date = now
        elif status in ['refused','failed']:
            task.status = 'pending'
            task.modification_date = now
            task.status_date = now
        elif status=='succeeded':
            task.status = 'succeeded'
            task.modification_date = now
            task.status_date = now
        else:
            raise ModelException(f"An execution cannot change status from pending to {status}")
    elif execution.status=='accepted':
        if status=='refused':
            task.stats = 'pending'
            task.modification_date = now
            task.status_date = now
        elif status in ['running','failed','succeeded','pending']:
            task.status = status
            task.modification_date = now
            task.status_date = now
        else:
            log.exception(f"An execution cannot change status from accepted to {status}")
            raise ModelException(f"An execution cannot change status from accepted to {status}")
    elif execution.status=='running':
        if status in ['succeeded', 'failed']:
            task.status=status
            task.modification_date = now
            task.status_date = now
            if status=='failed' and task.retry>0:
                log.warning(f'Failure of execution {execution.execution_id} trigger task {task.task_id} retry ({task.retry-1} retries left)')
                task.retry -= 1
                task.status = 'pending'
        else:
            log.exception(f"An execution cannot change status from running to {status}")
            raise ModelException(f"An execution cannot change status from running to {status}")
    else:
        raise ModelException(f"An execution cannot change status from {execution.status} (only from pending, running or accepted)")
    execution.status=status
    if commit:
        session.commit()


class Signal(db.Model):
    __tablename__ = "signal"
    #execution_id = db.Column(db.Integer, db.ForeignKey("execution.execution_id"), primary_key=True, nullable=True)
    signal_id = db.Column(db.Integer, primary_key=True)
    execution_id = db.Column(db.Integer, db.ForeignKey("execution.execution_id"), nullable=True)
    worker_id = db.Column(db.Integer, db.ForeignKey("worker.worker_id"))
    signal = db.Column(db.Integer, nullable=False)    
    execution = db.relationship(
        Execution,
        backref=db.backref('signals',
                            uselist=True,
                            cascade='delete,all'))

    def __init__(self, execution_id, worker_id, signal):
        self.execution_id = execution_id
        self.worker_id = worker_id
        self.signal = signal


class Job(db.Model):
    __tablename__ = "job"
    job_id = db.Column(db.Integer, primary_key=True)
    target = db.Column(db.String)
    action = db.Column(db.String)
    args = db.Column(db.JSON, default={})
    retry =  db.Column(db.Integer, default = 0)
    status = db.Column(db.String, default = 'pending')
    log = db.Column(db.Text, nullable=True)
    creation_date = db.Column(db.DateTime, server_default=func.now())
    modification_date = db.Column(db.DateTime, onupdate=func.now())
    

    def __init__(self, target, action, args={}, retry=0):
        self.target = target
        self.action = action
        self.args = args
        self.retry = retry

def worker_delete(worker, session, is_destroyed=False, commit=True):
    """Handle worker delete, take care of launching the destroy job if needed."""
    if not worker.permanent and not is_destroyed:
        create_worker_destroy_job(worker, session)
        return worker
    else:
        for execution in session.query(Execution).filter(Execution.worker_id==worker.worker_id, Execution.status=='running'):
            execution_update_status(execution, session, 'failed')
        for execution in session.query(Execution).filter(Execution.worker_id==worker.worker_id, Execution.status.in_(['pending','accepted'])):
            execution_update_status(execution, session, 'refused')
        session.delete(worker)
        session.commit()
        return worker

def create_worker_destroy_job(worker, session, commit=True):
    job = Job(target = worker.name,
        action='worker_destroy',
        args=to_dict(worker),
        retry=WORKER_DESTROY_RETRY)
    session.add(job)
    if commit:
        session.commit()

def create_worker_create_job(concurrency, prefetch, batch, flavor, region, provider, session, number=1, commit=True):
    for _ in range(number):
        session.add(
            Job(target='', 
                action='worker_create', 
                args={
                    'concurrency': concurrency, 
                    'prefetch': prefetch,
                    'flavor': flavor,
                    'region':region,
                    'provider':provider,
                    'batch':batch
                }
            )
        )
    if commit:
        session.commit()

def worker_handle_eviction(worker, session, commit=True):
    "Handle worker eviction"
    # Here we want to make an exception: Execution failure on worker eviction should be 
    # immediately retried whatever the retry status
    # NB could not make the SQLALchemy ORM work in that simple case... NotImplementedError: This backend does not support multiple-table criteria within UPDATE
    session.execute(f"UPDATE task SET status='pending' WHERE task_id IN (SELECT task_id FROM execution WHERE worker_id={worker.worker_id} AND status IN ('running','pending','accepted','assigned'))")
    session.execute(f"UPDATE execution SET status='failed' WHERE worker_id={worker.worker_id} AND status='running'")
    session.execute(f"UPDATE execution SET status='refused' WHERE worker_id={worker.worker_id} AND status IN ('pending','accepted')")
    
    if commit:
        session.commit()
    if EVICTION_ACTION in ['delete', 'replace']:
        create_worker_destroy_job(worker, session, commit=commit)
    if EVICTION_ACTION=='replace':
        pending_tasks=session.query(func.count(Task.task_id)).filter(Task.batch==worker.batch, Task.status=='pending').scalar()
        if pending_tasks > 0:
            flavor = worker.flavor_detail
            try:
                cost = session.query(FlavorMetrics.cost).filter(FlavorMetrics.flavor_name==flavor.name, 
                                                                FlavorMetrics.provider==flavor.provider,
                                                                FlavorMetrics.region_name==worker.region).one()
                cost = cost[0]
            except NoResultFound:
                cost = 0
            for new_flavor in find_flavor(session, 
                                        min_cpu=flavor.cpu, 
                                        min_ram=flavor.ram, 
                                        min_disk=flavor.disk, 
                                        limit=1000):
                if new_flavor["available"]==0:
                    continue
                if cost==0 and new_flavor['cost']==0:
                    break
                elif new_flavor['cost']/cost<EVICTION_COST_MARGIN:
                    break
            else:
                log.warning(f'Could not find any suitable replacement for flavor {flavor.name}')
                return None
            log.warning(f'Found a replacement flavor for {flavor.name}: {new_flavor["name"]} in {new_flavor["provider"]}:{new_flavor["region"]}')
            create_worker_create_job(concurrency=worker.concurrency, prefetch=worker.prefetch,
                                    batch=worker.batch, flavor=new_flavor["name"], region=new_flavor["region"],
                                    provider=new_flavor["provider"], session=session, commit=commit)
        




class Requirement(db.Model):
    __tablename__="requirement"
    requirement_id = db.Column(db.Integer, primary_key=True)
    task_id = db.Column(db.Integer, db.ForeignKey("task.task_id"), nullable=False)
    task = db.relationship(
        Task,
        foreign_keys=[task_id],
        backref=db.backref('requirements',
                         uselist=True,
                         cascade='delete,all'))
    other_task_id = db.Column(db.Integer, db.ForeignKey("task.task_id"), nullable=False)
    other_task = db.relationship(
        Task,
        foreign_keys=[other_task_id],
        backref=db.backref('required_in',
                         uselist=True,
                         cascade='delete,all'))

    def __init__(self, task_id, other_task_id):
        self.task_id=task_id
        self.other_task_id=other_task_id    
    
class Recruiter(db.Model):
    __tablename__="recruiter"
    batch = db.Column(db.String, default=DEFAULT_BATCH, nullable=False, primary_key=True)
    rank = db.Column(db.Integer, nullable=False, primary_key=True)
    tasks_per_worker = db.Column(db.Integer, nullable=False)
    worker_flavor = db.Column(db.String, nullable=False)
    worker_region = db.Column(db.String, nullable=True)
    worker_provider = db.Column(db.String, nullable=True)
    worker_concurrency = db.Column(db.Integer, nullable=False)
    worker_prefetch = db.Column(db.Integer, nullable=False, default=0)
    minimum_tasks = db.Column(db.Integer, nullable=True)
    maximum_workers = db.Column(db.Integer, nullable=True)

    def __init__(self, batch, rank, tasks_per_worker, 
                 worker_flavor, worker_concurrency,
                 worker_region = None, worker_provider = None, 
                 worker_prefetch = 0,
                 minimum_tasks=None, maximum_workers=None):
        self.batch = batch
        self.rank = rank
        self.tasks_per_worker = tasks_per_worker
        validate_protofilter(worker_flavor)
        self.worker_flavor = worker_flavor
        self.worker_region = worker_region
        self.worker_provider = worker_provider
        self.worker_concurrency = worker_concurrency
        self.worker_prefetch = worker_prefetch
        self.minimum_tasks = minimum_tasks
        self.maximum_workers = maximum_workers

    def match_flavor(self, worker, session):
        """A function that says if protofilters (which may appear as auto:... in Recruiter.worker_flavor) validate a worker.flavor"""
        if self.worker_flavor.startswith('auto'):
            if PROTOFILTER_SEPARATOR in self.worker_flavor:
                if worker.flavor_detail is None:
                    return False
                flavor_detail=worker.flavor_detail 
                env={'cpu':flavor_detail.cpu, 'ram':flavor_detail.ram, 'disk':flavor_detail.disk,'tags':flavor_detail.tags, 
                     'gpumem':flavor_detail.gpumem, 'region':worker.region,'provider':worker.provider}
                if 'cost' in self.worker_flavor or 'eviction' in self.worker_flavor:
                    try:
                        metrics = session.query(FlavorMetrics).filter(FlavorMetrics.flavor_name==worker.flavor,
                                                                    FlavorMetrics.provider==worker.provider,
                                                                    FlavorMetrics.region_name==worker.region).one()
                        env['cost']=metrics.cost
                        env['eviction']=metrics.eviction
                    except NoResultFound:
                        return False
                result = True
                for protofilter in self.worker_flavor.split(PROTOFILTER_SEPARATOR):
                    protofilter_match = protofilter_syntax.match(protofilter)
                    if protofilter_match:
                        protofilter_match = protofilter_match.groupdict()
                        variable = protofilter_match['item']
                        comp = protofilter_match['comparator']
                        value = protofilter_match['value']
                        if value[0] not in '0123456789.':
                            # value is a string
                            value=repr(value)
                        if comp=='~':
                            env['is_like']=is_like
                            eval_protofilter = f"is_like({variable},{value})"
                        elif comp=='#':
                            env['has_tag']=has_tag
                            eval_protofilter = f"has_tag({variable},{value})"
                        elif comp=='!~':
                            env['is_like']=is_like
                            eval_protofilter = f"not(is_like({variable},{value}))"
                        elif comp=='!#':
                            env['has_tag']=has_tag
                            eval_protofilter = f"not(has_tag({variable},{value}))"
                        else:
                            eval_protofilter = f"{variable}{comp}{value}"
                        try:
                            result = result and eval(eval_protofilter,env)
                        except Exception as e:
                            log.exception(f'Could not evaluate protofilter {protofilter} (compiled as {eval_protofilter}), eval to False')
                            return False
                return result
        else:
            return self.worker_flavor==worker.flavor
        
    # NB Session.merge() seems the way to go with this object
    # cf https://docs.sqlalchemy.org/en/14/orm/session_api.html#sqlalchemy.orm.Session.merge

def delete_batch(name, session, commit=True):
    """Delete a batch (all tasks, executions and recruiters associated to that batch), either in API context or in UI context"""
    session.execute(delete(Requirement).where(or_(
        Requirement.task_id.in_(select(Task.task_id).where(Task.batch==name)),
        Requirement.other_task_id.in_(select(Task.task_id).where(Task.batch==name)))),
        execution_options={'synchronize_session':False})
    session.execute(delete(Signal).where(Signal.execution_id.in_(
        select(Execution.execution_id).where(Execution.task_id.in_(
            select(Task.task_id).where(Task.batch==name))
        ))),
        execution_options={'synchronize_session':False})
    session.execute(delete(Execution).where(Execution.task_id.in_(
             select(Task.task_id).where(Task.batch==name))),
             execution_options={'synchronize_session':False})    
    session.execute(delete(Task).where(Task.batch==name),
             execution_options={'synchronize_session':False})
    session.execute(delete(Recruiter).where(Recruiter.batch==name),
             execution_options={'synchronize_session':False})
    if commit:
        session.commit()


class Region(db.Model):
    __tablename__="region"
    name = db.Column(db.String, nullable=False, primary_key=True)
    provider = db.Column(db.String, nullable=False, primary_key=True)
    
class Flavor(db.Model):
    __tablename__="flavor"
    name = db.Column(db.String, nullable=False, primary_key=True)
    provider = db.Column(db.String, nullable=False, primary_key=True)
    cpu = db.Column(db.Integer, nullable=False)
    ram = db.Column(db.Float, nullable=False)
    disk = db.Column(db.Float, nullable=False)
    bandwidth = db.Column(db.Float, nullable=True)
    gpu = db.Column(db.String, nullable=True)
    gpumem = db.Column(db.Float, nullable=True)
    tags = db.Column(db.String, nullable=True)
    workers = db.relationship('Worker', 
            primaryjoin=and_(name==Worker.flavor, provider==Worker.provider),
            foreign_keys=[Worker.flavor,Worker.provider],
            viewonly=True,
            backref='flavor_detail', 
            lazy=True)

class FlavorMetrics(db.Model):
    __tablename__='flavormetrics'
    flavor_name = db.Column(db.String, nullable=False, primary_key=True)
    provider = db.Column(db.String, nullable=False, primary_key=True)
    region_name = db.Column(db.String, nullable=False, primary_key=True)
    cost = db.Column(db.Float, nullable=False)
    eviction = db.Column(db.Integer, nullable=True)
    flavor = db.relationship('Flavor', 
            primaryjoin=and_(flavor_name==Flavor.name, provider==Flavor.provider),
            foreign_keys=[Flavor.name, Flavor.provider],
            viewonly=True,
            backref='metrics', 
            lazy=True)
    region = db.relationship('Region', 
            primaryjoin=and_(region_name==Region.name, provider==Flavor.provider), 
            foreign_keys=[Region.name, Region.provider],
            viewonly=True,
            backref='metrics', 
            lazy=True)

def find_remaining_quotas(session):
    """Return a dictionnary of (provider,region):cpus where cpus is the number of CPUs still available because of the quota"""
    quotas = get_quotas()
    if quotas:
        for provider,region,cpus in session.query(Worker.provider, Worker.region, func.sum(Flavor.cpu))\
                .select_from(Worker).join(Flavor,and_(Worker.flavor==Flavor.name,Worker.provider==Flavor.provider))\
                .group_by(Worker.provider, Worker.region):
            if (provider,region) in quotas:
                quotas[(provider,region)]-=cpus
        create_jobs=list(map(lambda x: x[0],session.query(Job.args).select_from(Job).filter(Job.action=='worker_create')))
        if create_jobs:
            flavors=[]
            for job_args in create_jobs:
                if (job_args['flavor'],job_args['provider']) not in flavors:
                    flavors.append((job_args['flavor'],job_args['provider']))
            flavors_cpu = { (name, provider):cpu for name, provider, cpu in
                session.query(Flavor.name, Flavor.provider, Flavor.cpu).select_from(Flavor).filter(
                    tuple_(Flavor.name, Flavor.provider).in_(flavors))
            }
            for job_args in create_jobs:
                quotas[(job_args['provider'],job_args['region'])]-=flavors_cpu.get((job_args['flavor'],job_args['provider']),0)
    return quotas    

def find_flavor(session, protofilters='', min_cpu=None, min_ram=None, min_disk=None, 
                max_eviction=FLAVOR_DEFAULT_EVICTION, 
                limit=FLAVOR_DEFAULT_LIMIT, provider=None, region=None, flavor=None):
    """Return a list of Flavor fulfilling specific conditions

    protofilters should be a string of PROTOFILTER_SEPARATOR (e.g. :) separated strings passing validate_protofilter() function
    """
    # WATCH OUT When modifying this function, Recruiter.match_flavor() function (used for recycling) should be updated too
    filters = []
    order_by = (FlavorMetrics.cost,)
    if max_eviction is not None and not (protofilters and 'eviction' in protofilters):
        filters.append(FlavorMetrics.eviction<=max_eviction)
    if not (protofilters and 'tags' in protofilters):
        filters.append(~Flavor.tags.like('%M%'))
    if min_cpu is not None:
        filters.append(Flavor.cpu>=min_cpu)
    if min_ram is not None:
        filters.append(Flavor.ram>=min_ram)
    if min_disk is not None:
        filters.append(Flavor.disk>=min_disk)
    if provider is not None:
        filters.append(Flavor.provider==provider)
        if provider in PREFERRED_REGIONS and PREFERRED_REGIONS[provider]:
            order_by = (FlavorMetrics.region_name.like(PREFERRED_REGIONS[provider])).desc(),*order_by
    if region is not None:
        filters.append(FlavorMetrics.region_name==region)
    if flavor is not None:
        filters.append(Flavor.name.like(flavor))
    if protofilters:
        for protofilter_string in protofilters.split(PROTOFILTER_SEPARATOR):
            protofilter = protofilter_syntax.match(protofilter_string)
            if protofilter:
                protofilter=protofilter.groupdict()
                protofilter_item = getattr(FlavorMetrics, 'region_name' if protofilter['item']=='region' else protofilter['item']) \
                    if protofilter['item'] in ['region','cost','eviction'] else \
                    getattr(Flavor, protofilter['item'])
                comp = protofilter['comparator']
                if comp=='==':
                    filters.append(protofilter_item==protofilter['value'])
                elif comp=='!=':
                    filters.append(protofilter_item!=protofilter['value'])
                elif comp=='>':
                    filters.append(protofilter_item>protofilter['value'])
                elif comp=='<':
                    filters.append(protofilter_item<protofilter['value'])
                elif comp=='>=':
                    filters.append(protofilter_item>=protofilter['value'])
                elif comp=='<=':
                    filters.append(protofilter_item<=protofilter['value'])
                elif comp=='~':
                    filters.append(protofilter_item.like(protofilter['value']))
                elif comp=='!~':
                    filters.append(~protofilter_item.like(protofilter['value']))
                elif comp=='#':
                    for letter in protofilter['value']:
                        filters.append(protofilter_item.like(f'%{letter}%'))
                elif comp=='!#':
                    for letter in protofilter['value']:
                        filters.append(~protofilter_item.like(f'%{letter}%'))
            else:
                log.warning(f'Unknown protofilter {protofilter_string}')
    #return session.query(Flavor,FlavorMetrics
    fields = ['name','provider','region','cpu','ram','tags','gpu','gpumem','disk','cost','eviction']
    flavors = [dict(zip(fields, object)) for object in session.query(
                  Flavor.name, Flavor.provider, FlavorMetrics.region_name, 
                  Flavor.cpu, Flavor.ram, Flavor.tags, Flavor.gpu, Flavor.gpumem, 
                  Flavor.disk, FlavorMetrics.cost, FlavorMetrics.eviction      
            ).select_from(FlavorMetrics).join(FlavorMetrics.flavor)\
                .filter(*filters)\
                .order_by(*order_by).limit(limit)]
    quotas = find_remaining_quotas(session)
    for flavor in flavors:
        if (flavor['provider'], flavor['region']) in quotas:
            flavor['available']=quotas[(flavor['provider'], flavor['region'])]//flavor['cpu']
        else:
            flavor['available']=None
    return flavors

class FlavorStats(db.Model):
    __tablename__='flavorstats'
    flavor_name = db.Column(db.String, nullable=False, primary_key=True)
    provider = db.Column(db.String, nullable=False, primary_key=True)
    region_name = db.Column(db.String, nullable=False, primary_key=True)
    rank = db.Column(db.Integer, nullable=False, primary_key=True)
    failure_rate = db.Column(db.Float, default=0)
    eviction_rate = db.Column(db.Float, default=0)
    event_number = db.Column(db.Integer, default=0)
    flavor = db.relationship('Flavor', 
            primaryjoin=and_(flavor_name==Flavor.name, provider==Flavor.provider),
            foreign_keys=[Flavor.name, Flavor.provider],
            viewonly=True,
            backref='stats', 
            lazy=True)
    region = db.relationship('Region', 
            primaryjoin=and_(region_name==Region.name, provider==Flavor.provider), 
            foreign_keys=[Region.name, Region.provider],
            viewonly=True,
            backref='stats', 
            lazy=True)

    def deploy_success(self):
        self.failure_rate = self.failure_rate * self.event_number / (self.event_number + 1)
        self.event_number+=1

    def deploy_failure(self):
        self.failure_rate = (self.failure_rate * self.event_number + 1) / (self.event_number + 1)
        self.event_number+=1

    def eviction_event(self):
        self.eviction_rate = (self.eviction_rate * (self.event_number-1) + 1) / self.event_number

    def destroy_success(self):
        self.eviction_rate = (self.eviction_rate * (self.event_number-1) ) / self.event_number



"""Integration of ansible inventory in main database

Revision ID: 2be1e584707d
Revises: 740850a52979
Create Date: 2024-07-01 19:07:11.597036

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2be1e584707d'
down_revision = '740850a52979'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.add_column(sa.Column('ipv4', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('ipv6', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('ansible_host', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('ansible_group', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('ansible_active', sa.Boolean(), server_default="0"))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('worker', schema=None) as batch_op:
        batch_op.drop_column('ansible_active')
        batch_op.drop_column('ansible_group')
        batch_op.drop_column('ansible_host')
        batch_op.drop_column('ipv6')
        batch_op.drop_column('ipv4')

    # ### end Alembic commands ###



from scitq.workflow import Workflow

#wf = Workflow(name='wf1', shell=True, workers_per_task=2, total_workers=1, retry=2, container='alpine', flavor='local')
wf = Workflow(name='wf1', shell=True, workers_per_task=2, retry=2, container='alpine', provider='azure', region='swedencentral', flavor='Standard_D2ads_v5')

for i in range(100):
    s1 = wf.step(
        batch='s1',
        name=f's1:{i}',
        command='sleep 10',
        concurrency=10,
    )
    s2 = wf.step(
        batch='s2',
        name=f's2:{i}',
        command='sleep 20',
        concurrency=5,
        required_tasks=s1,
    )

s3 = wf.step(
    batch='s3',
    name='final',
    command='echo "Done!"',
    required_tasks=s2.gather(),
    concurrency=1
)

wf.run(refresh=1)
print(f'Result is {s3.get_output()}')
wf.clean()





import json
import requests
from time import sleep
from .lib import Server
import logging as log

AZURE_EVENT_URL ="http://169.254.169.254/metadata/scheduledevents"
AZURE_HEADERS = {'Metadata' : 'true'}
AZURE_PARAMS = {'api-version':'2020-07-01'}

WAITING_TIME = 2

def get_azure_events():           
    resp = requests.get(AZURE_EVENT_URL, headers = AZURE_HEADERS, params = AZURE_PARAMS)
    data = resp.json()
    events = data.get('Events',[])
    if events and events[0].get('EventType',None)=='Preempt':
        return True
    else:
        return False

def monitor_events(server, provider, worker_id):
    """Thread to launch when the client is launched to monitor events"""
    if provider=='azure':
        log.warning(f'{provider} eviction monitor started')
        get_event = get_azure_events
    else:
        log.warning(f'No events to monitor for provider {provider}')
        return None
    s = Server(server, style='object')
    while True:
        if get_event():
            log.exception('An eviction event was detected.')
            s.worker_update(worker_id,status='evicted', asynchronous=False)
            return None
        sleep(WAITING_TIME)           

